<?php
namespace ID3\Blog;

class Synchroniser{
    public $settings = null;
    public $logHandler = null;
    private $_locale = null;
    
    public $state = [
        'message' => 'idle',
        'current_index' => 0,
        'total' => 0
    ];

    private $categoryReferenceTable = [];

    function __construct(){
        $this->logHandler = new \ID3\Bases\PluginLogger(KEY);

    }
    static $_instance = null;
    static function getInstance(){
        if(self::$_instance == null ){
            self::$_instance = new Synchroniser();
        }

        return self::$_instance;
    }

    function loadSettings(){
        $this->settings = \ID3\Settings::get(KEY, [
            'force_sync_timeout' => 5
        ]);

        
        $this->updateSettings();
    }

    function clear(){
        $result = [
            'deleted_articles' => $this->clearArticles(),
            'deleted_categories' => $this->clearCategories(),
        ];

        $this->logHandler->write('Clear result');
        $this->logHandler->write($result);
        return $result;
    }
    
    function clearArticles($categories=null){
        $result = 0;
        
        $this->logHandler->write('ClearArticles BEGINS');

        $delete_post = array(
            'posts_per_page' => -1,    // The number of posts to retrieve, otherwise 5
            'post_type'   => 'post',
            'post_status' => array('future','publish'),
            'meta_query' => array(
                array(
                    'key' => 'reference_id',
                    'compare' => 'EXISTS'
                )
            )
        );

        if( $categories !== null && is_array($categories)){
            $delete_post['category__in'] = $categories;
        }

        if($categories === null){
            $this->logHandler->write('$categories is NULL');
            $delete_post['suppress_filters'] = true;
        }

        $q = new \WP_Query($delete_post);
        $deletedPosts = [];

        foreach ($q->posts as $post) {
            $attachments = get_children($post->ID);

            foreach ($attachments as $attachment) {
                wp_delete_attachment( $attachment->ID, true );
            }

            $deleteResult = wp_delete_post($post->ID, true);
            if($deleteResult !== false || $deleteResult !== null){
                $deletedPosts[] = $post->ID;
                $result++;
            }

        }
        $this->logHandler->write('query:');
        $this->logHandler->write($delete_post);
        $this->logHandler->write('deleted:');
        $this->logHandler->write($deletedPosts);
        $this->logHandler->write('ClearArticles END');
        return $result;
    }
    function clearCategories(){
        $result = 0;

        $args = [
            'hide_empty' => false, // also retrieve terms which are not used yet
            'suppress_filters' => true,
            'meta_query' => [
                [
                   'key'       => 'reference_id',
                   'compare'   => 'EXISTS'
                ]
            ],
            'taxonomy'  => 'category',
        ];

        global $sitepress;
        if ( $sitepress ) {
            remove_filter( 'get_terms_args', array( $sitepress, 'get_terms_args_filter' ) );
            remove_filter( 'get_term', array( $sitepress, 'get_term_adjust_id' ) );
            remove_filter( 'terms_clauses', array( $sitepress, 'terms_clauses' ) );
            $terms = get_terms( $args );
            add_filter( 'terms_clauses', array( $sitepress, 'terms_clauses' ), 10, 4 );
            add_filter( 'get_term', array( $sitepress, 'get_term_adjust_id' ), 1, 1 );
            add_filter( 'get_terms_args', array( $sitepress, 'get_terms_args_filter' ), 10, 2 );
        }
        else {
            $terms = get_terms( $args );
        }

        //$this->logHandler->write($result);
        foreach ($terms as $cat) {
            delete_term_meta($cat->term_id,'reference_id');
            $deleteResult = wp_delete_term($cat->term_id,'category');
            if($deleteResult !== false || $deleteResult !== null){
                $result++;
            }
        }
        
        return $result;
    }

    function start($options=[]){
        $options = array_merge([
            'check_pending'     => false,
            'update_categories' => true,
            'update_articles'   => true,
            'method'            => 'incremental',
            'date'              => null
        ],$options);


        $this->loadSettings();

        $result = [
            'new' => 0,
            'update' => 0,
            'delete' => 0,
            'error' => 0,
        ];

        // When last_sync is older than a the force_sync_timeout settings (in days), skip the check pending
        if($this->settings->last_sync != null && $this->settings->last_sync < strtotime("-{$this->settings->force_sync_timeout} days")){
            $options['check_pending'] = false;
        }

        if($options['check_pending'] == true){
            if(!$this->hasPendingUpdate()){
                $this->clearPendingUpdate();
                return $result;
            }
        }

        $endpoint = $this->getSynchronisationEndpoint();
        
        if($endpoint === null) {
            $result['message'] = "Can't get a valid Endpoint";
            $result['settings'] = $this->settings;
            return $result;
        }

        $languages = $this->getRemoteActiveLocales($endpoint);
        //$this->logHandler->write($languages);
        error_log('Syncing for ' . json_encode($languages) . ' started');

        foreach ($languages as $locale) {
            
            $this->switchLocale($locale);

            if($options['update_categories']){
                $this->logHandler->write("Sync {$locale} categories");

                $syncResult = $this->syncCategories($endpoint,$locale,$result);
                $result = $this->sumArrays($result, $syncResult);
                $result['categories'] = $syncResult;
            }
    
            if($options['method'] === 'full'){
                // When the sync method is set to full, remove local article that don't fit the current exceptions    
                if(isset($this->settings->category_exceptions) && !empty($this->settings->category_exceptions->{$locale}->list)){
                    $excluded_cats = [];
                    foreach ($this->settings->category_exceptions->{$locale}->list as $cat) {
                        $excluded_cats[] = $cat->key;
                    }

    
                    $result['delete'] = $this->clearArticles($this->convertRemoteCategoryToLocal($excluded_cats));
                }
            }
    
            if($options['update_articles']){

                $this->logHandler->write("Sync {$locale} articles");
                
                $syncResult = $this->syncArticles($endpoint,$locale, $options,$result);
                $result = $this->sumArrays($result, $syncResult);
                $result['articles'] = $syncResult;
            }
            $this->logHandler->write("Sync {$locale} RESULTS:");
            $this->logHandler->write($result);
        }

        $this->resetLocale();
        
        $this->resetState();

        $result['last_sync_saved'] = $this->updateTimestamp();
        $this->clearPendingUpdate();

        if($result['new'] > 0 || $result['update'] > 0 || $result['delete'] > 0){
            \do_action('id3-blog/sync/new-articles');
        }

        return $result;
    }

    function switchLocale($lang){
        if ( \ID3\is_plugin_active('sitepress') ) {
            if($this->_locale == null){
                $this->_locale = apply_filters( 'wpml_current_language', null );
            }

            do_action( 'wpml_switch_language', $lang );

            if($lang != apply_filters( 'wpml_current_language', null )){
				global $sitepress;
				$sitepress->switch_lang($lang);
			}
			
			   
			if($lang != apply_filters( 'wpml_current_language', null )){
				error_log('failed to switch locale to ' . $lang);	
                return false;
			}

            return true;
        }
        

        return false;
    }
    function resetLocale(){
        if($this->_locale != null){
            do_action( 'wpml_switch_language', $this->_locale );
            $this->_locale = null;
        }
    }

    function resetState(){
        $this->state['message'] = 'idle';
        $this->state['total'] = 0;
        $this->state['current_index'] = 0;
    }

    function updateTimestamp(){
        $this->loadSettings();
        $this->settings->last_sync = time();


        return \ID3\Settings::set(KEY, $this->settings);
    }
    

    function hasPendingUpdate(){
        $timestamp = get_transient(KEY . '-pending-update');
        if($timestamp === false) return false;
        if($this->settings->last_sync == null) return true;

        return $timestamp > $this->settings->last_sync;
    }

    function clearPendingUpdate(){
        delete_transient(KEY . '-pending-update');
    }

    function getRemoteActiveLocales($endpoint){
        $result = \ID3\Helpers\remote_api_call($endpoint.'/wp-json/id3-dlp/supported-languages');
        $result = json_decode($result);

        if(!is_array($result)){
            return ['fr'];
        }

        return $result;
    }

    function syncCategories($endpoint, $locale, &$globalResults){
        $new = 0;
        $update = 0;
        $error = 0;
        $cats = [];

        $this->state['message'] = 'getting categories';
        $categoriesToUpdate = [];
        $categoriesToUpdate = $this->getRemoteCategories($endpoint, $locale);

        $this->logHandler->write('syncing categories');
        $this->logHandler->write($categoriesToUpdate);

        if(!is_array($categoriesToUpdate) && isset($categoriesToUpdate['error'])){
            return [
                'error_message' => $categoriesToUpdate['error']
            ];
        }
        $this->state['message'] = 'syncing categories';
        $this->state['total_categories'] = count($categoriesToUpdate);

        foreach ($categoriesToUpdate as $cat) {
            if(in_array($cat->name, ['Autre','Other','Uncategorized',''])) continue;

            $updateResult = $this->updateCategory($cat);
            
            if($updateResult['success']===true){    
                //$this->logHandler->write('newOrUpdate category: ' . $cat->name . ' successfull');
                if($updateResult['newOrUpdate'] == 'new') $new++;
                if($updateResult['newOrUpdate'] == 'update') $update++;
            }
            else{
               $error++;
               $globalResults['error_message'] = $updateResult['error'];
               $this->logHandler->write($cat->name . ' failed to update or create');
            }
        }

        
        return [
            'total'=> count($categoriesToUpdate),
            'new' => $new,
            'update' => $update,
            'error' => $error
        ];
    }

    function syncArticles($endpoint, $locale, $options, &$globalResults){
        $new = 0;
        $update = 0;
        $delete = 0;
        $error = 0;
        $updates = [];


        $this->state['message'] = 'getting articles';
        $lastSync =  '2021-01-01 00:00:00';
        if(isset($this->settings->last_sync) && $this->settings->last_sync != null){
            if(is_numeric($this->settings->last_sync)) $lastSync = date("Y-m-d H:i:s", $this->settings->last_sync);
            else $lastSync = date_format($this->settings->last_sync,'Y-m-d H:n:s');
        }
        
        $articlesToUpdate = [];



        $filters = $this->getRemoteArticlesFilters($locale, $this->settings->category_subscriptions, $this->settings->category_exceptions);
        

        $limitDate = $options['date'] !== null ? $options['date'] : $lastSync;
        $articlesToUpdate = $this->getRemoteArticles($endpoint,$locale, $limitDate, $filters, $options['method']);

        $this->state['message'] = 'syncing articles';
        $this->state['total_articles'] = 0;

        if($articlesToUpdate != null){
            $this->state['total_articles'] = count($articlesToUpdate);
            $this->state['current_index'] = 0;

            $this->logHandler->write('syncing ' . $locale . ' articles');
            $this->logHandler->write($filters);
            //$this->logHandler->write($articlesToUpdate);

            foreach ($articlesToUpdate as $article) {
                $this->state['current_index'] +=1;
                $updateResult = $this->updateArticle($article);

                if($updateResult['success']===true){    
                    if($updateResult['newOrUpdate'] == 'new') $new++;
                    if($updateResult['newOrUpdate'] == 'update') $update++;
                    $this->logHandler->write($updateResult['newOrUpdate'] . ' article: ' . $article->post_title);
                    $updates[] = $updateResult; //['post_categories'];
                }
                else{
                    $this->logHandler->write(['failed' => $article]);
                $error++;
                }
            }
        }
        $result = [
            'total'=> $this->state['total_articles'],
            'new' => $new,
            'update' => $update,
            'delete' => $delete,
            'updates' => $updates,
            'error' => $error,
            'from_date' => $limitDate
        ];
        $this->logHandler->write($result);
        

        return $result;
    }


    function updateCategory($newCategoryData){
        $result = [
            'success' => false,
            'newOrUpdate' => 'new',
            'id' => 0
        ];
        // backup reference ID and remove it from the object
        // this ID is use to cross reference the local and remote article ID
        $ref_id = $newCategoryData->cat_ID;
        $keyToUpdate = ['name','description','nicename'];
        $update_result = false;

        // try to get a stored version of the article
        $localCategory = $this->getStoredCategory($ref_id);
        if($localCategory == null){
            $update_result = wp_insert_term($newCategoryData->name,'category',[
                'description' => $newCategoryData->description
            ]);
        }
        else{
            $result['newOrUpdate'] = 'update';
            $update_result = wp_update_term($localCategory->term_id,'category',[
                'name' => $newCategoryData->name,
                'description' => $newCategoryData->description
            ]);
        }
        
        if(!is_wp_error($update_result)){
            $result['success'] = true;
            $result['id'] = $tid = $update_result['term_id'];

            $this->addCategoryReference($ref_id, $tid);
            if($result['newOrUpdate'] == 'new'){
                // add reference_id meta
                add_metadata('term',$tid, 'reference_id', $ref_id, true);
                $locale = isset($newCategoryData->lang) ? $newCategoryData->lang : null;
                $this->setCategoryLanguage($tid, $locale);
            }  
        }
        else{
            $result['error'] = $update_result->errors;
            //return $update_result;
        }

        return $result;
    }

    function updateArticle($newArticleData){
        $result = [
            'success' => false,
            'newOrUpdate' => 'new',
            'id' => 0,
            'post_categories' => 'none'
        ];
        // backup reference ID and remove it from the object
        // this ID is use to cross reference the local and remote article ID
        $ref_id = $newArticleData->ID;
        
        // try to get a stored version of the article
        $localArticle = $this->getStoredArticle($ref_id);
        if($localArticle == null){
            // this article doesn't exists locally
            // and must be save to get and ID
            $localArticle = $newArticleData;
            // set ID to 0 to make sure it will insert a new article
            $localArticle->ID = 0;
        }
        else{
            // if($localArticle->post_modified > $newArticleData->post_modified){
            //     $result['newOrUpdate'] = 'skipped';
            //     return $result;
            // }
            // Copy data to localArticle
            $this->copyToObject($newArticleData, $localArticle, ['post_title','post_content','post_excerpt','post_name','post_date','post_date_gmt','post_modified','post_modified_gmt','post_type']);
            $result['newOrUpdate'] = 'update';
        }

        
        $localArticle->post_author = $this->settings->article_author;

        // add meta input
        $localArticle->meta_input = [
            'reference_id' => $ref_id,
            //'lang' => $newArticleData->lang || $defaultLocale
        ];
        

        // insert (if $localArticle->ID is NOT 0, the method acts the as wp_update_post)
        $insert_result = wp_insert_post($localArticle, true);

        if(!is_wp_error($insert_result)){
            $result['success'] = true;
            $result['attachment'] = null;
            $result['id'] = $pid = $insert_result;

            if(isset($newArticleData->categories)){
                $result['post_categories'] = $this->convertRemoteCategoryToLocal($newArticleData->categories);
                wp_set_post_categories( $pid, $result['post_categories'] , false );
            }
            if(isset($newArticleData->feature_image)){
                $result['attachment'] = $this->addArticleFeatureImage($pid, $newArticleData->feature_image);
            }

            if(isset($newArticleData->metas)){
                foreach($newArticleData->metas as $meta){
                    update_post_meta($pid, $meta->key, $meta->value);
                }
            }

            $this->setPostLanguage($pid, $newArticleData->lang);
        }

        return $result;
    }

    function addArticleFeatureImage($post_id, $image_url){
        if(strpos($image_url,'http')===false){
            // This should be a remote image, the lack of absolute address is explained by relay process
            // Eitherway, add the absolute path
            
        }
        $filepath = $image_url;
        $filesizes = getimagesize($filepath);
        $filename = pathinfo($filepath)['filename'];

        $result = [];

        // Get already use filename id, if any        
        $attachment_id = $this->getExistingAttachments($filename);

        // If there's no attachment with this filename, add it
        if($attachment_id == null){
            $attachment_id = wp_insert_attachment([
                'post_mime_type' => $filesizes['mime'],
                'post_title' => $filename,
                'post_content' => $filepath,
                'post_status' => 'inherit'
            ],$filepath, $post_id);
    
            $result['attachment'] = $attachment_id;
        }
        
        //$result['deleted_attachments'] = $deletedAttachments;

        if(!is_wp_error($attachment_id)){

            $attachment_data = [
                'width' => $filesizes[0],
                'height' => $filesizes[1],
                'file'  => $filepath
            ];


            wp_update_attachment_metadata( $attachment_id,  $attachment_data );
            update_post_meta( $post_id, '_thumbnail_id', $attachment_id );

        }

        return $result;
    }

    function clearExistingAttachments($filename){
        global $wpdb;

        $attPostIds = $wpdb->get_col("select ID from $wpdb->posts where post_type='attachment' AND post_title = '".$filename."' ");

        foreach ($attPostIds as $id){
            wp_delete_post($id, true);
        }

        return [$filename, count($attPostIds)];
    }

    function getExistingAttachments($filename){
        global $wpdb;

        $attPostIds = $wpdb->get_col("select ID from $wpdb->posts where post_type='attachment' AND post_title = '".$filename."' ");
        if(count($attPostIds) > 0){
            return $attPostIds[0];
        }
        return null;
    }

    function getStoredCategory($reference_id){
        $args = [
            'hide_empty' => false, // also retrieve terms which are not used yet
            'meta_query' => [
                [
                   'key'       => 'reference_id',
                   'value'     => $reference_id,
                   'compare'   => '='
                ]
            ],
            'taxonomy'  => 'category',
        ];

        if ( \ID3\is_plugin_active('sitepress') ) {
            global $sitepress;
            remove_filter( 'terms_clauses', array( $sitepress, 'terms_clauses' ) );
        }        

        $terms = get_terms($args);
        if(count($terms) > 0){
            return $terms[0];
        }
        return null;
    }

    function getStoredArticle($reference_id){
        $queryOptions = [
            'posts_per_page' => -1,
            'post_type' => 'post',
            'post_status' => ['publish','future'],
            'meta_key' => 'reference_id',
            'meta_value' => $reference_id,
            'suppress_filters' => true,
        ];

        $q = new \WP_Query( $queryOptions );
        if(count($q->posts) > 0){
            return $q->posts[0];
        }
        return null;
    }


    function getRemoteCategories($endpoint, $locale){
        $result = [];
        $result = \ID3\Helpers\remote_api_call($endpoint.'/wp-json/id3-dlp/categories', ['locale' => $locale]);
        
        if(isset($result['error'])){
            return $result;
        }
        
        $result = json_decode($result);
        return $result;
    }

    // Get Article list from remote source
    // Those articles creation or update must be more recent than the 
    // provided date 
    function getRemoteArticles($endpoint,$locale, $fromDate, $filters, $method) {
        $result = [];
        
        $json_data = json_encode([
            'locale' => $locale,
            'slave_id' => $_SERVER['SERVER_NAME'],
            'limit_date' => $fromDate == null ? '2021-01-01' : $fromDate,
            'filters' => $filters,
            'sync_method' => $method
        ]);

        $this->logHandler->write('getRemoteArticles/remote_api_call(' .  json_encode($json_data) . ') @ ' . $endpoint.'/wp-json/id3-dlp/salvage');

        $result = \ID3\Helpers\remote_api_call($endpoint.'/wp-json/id3-dlp/salvage', $json_data, 'POST');

        $result = json_decode($result);
        return $result->articles;
    }

    function getRemoteArticlesFilters($locale, $filter_in, $filter_out){
        $result = [];
       
        $list = isset($filter_in->{$locale}->list) ? $filter_in->{$locale}->list : [];
        $in = json_decode(json_encode([
            "type" => "subscriptions",
            "all" => $filter_in->{$locale}->all,
            "list" => array_map(
                        function($item){
                            return $item->key;
                        },
                        $list
                    )
        ]));

        $result[] = $in;

        if($filter_out != null){
            $list = isset($filter_out->{$locale}->list) ? $filter_out->{$locale}->list : [];
            $out = json_decode(json_encode([
                "type" => "exceptions",
                "all" => false,
                "list" => array_map(
                            function($item){
                                return $item->key;
                            },
                            $list
                        )
            ]));
            $result[] = $out; 
        }
        //$this->logHandler->write('getRemoteArticlesFilters(' . $locale . ')' . json_encode($in) . ' , ' . json_encode($out));

        return $result;
    }

    function getLocalArticlesFilters($filter_in, $filter_out){
        $result = [];

        $in = json_decode(json_encode($filter_in));
        $in->list = array_map(function($item){
            $cat_id = $item->key;
            if(isset($this->categoryReferenceTable[$cat_id])){
                $cat_id = $this->categoryReferenceTable[$cat_id];
            }
            return $cat_id;
        },$filter_in->list);
        $in->type = 'subscriptions';
        $result[] = $in;

        if($filter_out != null){
            $out = json_decode(json_encode($filter_out));
            $out->list = array_map(function($item){
                $cat_id = $item->key;
                if(isset($this->categoryReferenceTable[$cat_id])){
                    $cat_id = $this->categoryReferenceTable[$cat_id];
                }
                return $cat_id;
            },$filter_out->list);
            $out->type = 'exceptions';
            $result[] = $out; 
        }

        return $result;
    }

    function getSynchronisationEndpoint(){
        if($this->settings === null) return null;

        if(strToUpper($this->settings->source) === 'CUSTOM'){
            return ($this->settings->source_custom === '') ? null : $this->settings->source_custom;
        }

        return ($this->settings->source === '') ?  null : 'https://' . $this->settings->source;
    }

    function copyToObject($from, &$to, $keys = []){

        foreach ($keys as $key) {
            $to->{$key} = $from->{$key};
        }
    }

    function sumArrays($array1, $array2){
        $new = [];
        foreach($array1 as $key => $value){
            if(!is_array($value)){
                $new[$key] = $array1[$key] + (isset($array2[$key]) ? $array2[$key] : 0);
            }
            else{
                $new[$key] = $array1[$key];
            }
            
        }
        return $new;
    }

    function addCategoryReference($remoteId, $localId){
        $this->categoryReferenceTable[$remoteId] = $localId;
    }

    function convertRemoteCategoryToLocal(Array $remoteIds){
        $result = [];
        foreach ($remoteIds as $cat_id) {
            if(isset($this->categoryReferenceTable[$cat_id])){
                $result[] = $this->categoryReferenceTable[$cat_id];
            }
        }
        return $result;
    }


    /*
    Localization management
    */
    function setPostLanguage($post_id, $lang){

        if($lang==''){
            $lang = $this->getDefaultLocale();
        }

        // Polylang
        if(function_exists('pll_set_post_language')){
            \pll_set_post_language($post_id, $lang);
            return;
        }

        // WPML
        if(\ID3\is_plugin_active('sitepress')){
            $type = \get_post_type( $post_id );
 
        
            // Update the post language info
            $language_args = [
                'element_id' => $post_id,
                'element_type' => 'post_'.$type,
                'trid' => false,
                'language_code' => $lang,
                'source_language_code' => null,
            ];
        
            \do_action( 'wpml_set_element_language_details', $language_args );

            return;
        }
        
        update_post_meta($post_id, 'lang', $lang);
    }
    function setCategoryLanguage($term_id, $lang){

        if($lang==''){
            $lang = $this->getDefaultLocale();
        }

        // Polylang
        if(function_exists('pll_set_term_language')){
            pll_set_term_language($term_id, $lang);
            return;
        }

        if(\ID3\is_plugin_active('sitepress')){
            
            // Update the post language info
            $language_args = [
                'element_id' => $term_id,
                'element_type' => 'tax_category',
                'trid' => false,
                'language_code' => $lang,
                'source_language_code' => null,
            ];
        
            \do_action( 'wpml_set_element_language_details', $language_args );

            return;
        }

        add_metadata('term',$term_id, 'lang', $lang, true);
    }

    function getDefaultLocale(){
        if(function_exists('pll_default_language')){
            return pll_default_language();
        }

        $defaultLang = get_locale();
        $defaultLang = apply_filters('wpml_default_language', $defaultLang );

        return $defaultLang;

    }


    function updateSettings(){
        $settingChanged = false;

        // check filters
        if(isset($this->settings->category_subscriptions->all)){
            // error_log('Setting are not up to date:');
            // error_log(json_encode($this->settings));


            // Subscriptions 
            $this->settings->category_subscriptions = json_decode(json_encode(
                [
                    'fr' => [
                        'all' => $this->settings->category_subscriptions->all,
                        'list' => $this->settings->category_subscriptions->list
                    ]
                ]
            ));
        
            // Exceptions
            $this->settings->category_exceptions = json_decode(json_encode(
                [
                    'fr' => [
                        'all' => $this->settings->category_exceptions->all,
                        'list' => $this->settings->category_exceptions->list
                    ]
                ]
            ));

            
            $settingChanged = true;
        }

        if(!isset($this->settings->article_author)){
            $userList = \get_users();
            $this->settings->article_author = $userList[0]->ID;
            
            
            $settingChanged = true;
        }


        if($settingChanged){
            \ID3\Settings::set(KEY, $this->settings);
        }
        // error_log('Setting now:');
        // error_log(json_encode($this->settings));

    }

}